# encoding: utf-8

# Exemplo de implementação de árvore binária de busca.
#
# Página: http://malbarbo.pro.br/ensino/2013/7256/
# Autor : Marco A L Barbosa <marco@malbarbo.pro.br>

from aed import Struct, executa_testes

__all__ = ['ArvoreBB']

class No(Struct):
    '''
    Representa um nó em uma árvore binária
        valor : Qualquer - representa o valor no nó
        esq : No - representa a subárvore a esquerda
        dir : No - representa a subárvore a direita
    '''
    def __init__(self, valor, esq, dir):
        self.init(valor, esq, dir)


def altura_no(r):
    '''
    No -> Natural
    Devolve a altura da subárvore com raiz r.
    Veja os exemplos d método altura da classe ArvoreBB.
    '''
    if r == None:
        return -1
    else:
        aesq = altura_no(r.esq)
        adir = altura_no(r.dir)
    return max(aesq, adir) + 1


def tamanho_no(r):
    '''
    ArvoreBB, No -> Natural
    Devolve a quantidade de nós na subávore com raiz r.
    Veja os exemplos do método __len__ da classe ArvoreBB.
    '''
    if r == None:
        return 0
    else:
        tesq = tamanho_no(r.esq)
        tdir = tamanho_no(r.dir)
        return tesq + tdir + 1


def cria_arvore_teste():
    r'''
    -> ArvoreBB
    Cria uma árvore para ser usada nos testes
           10
         /    \
        5      15
      /   \    /
     3     6  12
      \
       4

    '''
    t = ArvoreBB()
    t.insere(10)
    t.insere(5)
    t.insere(6)
    t.insere(15)
    t.insere(3)
    t.insere(4)
    t.insere(12)
    return t


class ArvoreBB(Struct):
    '''
    Representa uma árvore binária de busca (ABB). Uma ABB é uma árvore com a
    seguinte propriedade:
        Seja x uma nó na ABB. Se y é um nó na subárvore a esquerda de x, então
        y.valor <= x.valor. Se y é uma nó na subárvore a direira de x, então
        y.valor >= x.valor
    Campos
        _raiz : No - representa a raiz da árvore
    '''
    def __init__(self):
        self.init()
        self._raiz = None

    def _busca(self, v):
        '''
        ArvoreBB, Qualquer -> No
        Se v está na árvore, devolve o nó que contém v, caso contrário devolve None.
        Veja os exemplos de __contains__.
        '''
        p = self._raiz
        while p != None:
            if v == p.valor:
                return p
            elif v < p.valor:
                p = p.esq
            else:
                p = p.dir
        return None

    def __contains__(self, v):
        '''
        ArvoreBB, Qualquer -> Boolean
        Devolve True se v está na árvore, False caso contrário.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> 5 in t
        True
        >>> 10 in t
        True
        >>> 4 in t
        True
        >>> 12 in t
        True
        >>> 17 in t
        False
        >>> 0 in t
        False
        '''
        p = self._busca(v)
        return p != None

    def insere(self, v):
        '''
        ArvoreBB, Qualquer -> None
        Insere v na árvore.
        >>> t = ArvoreBB()
        >>> t.insere(8)
        >>> t._raiz
        No(8, None, None)
        >>> t.insere(4)
        >>> t._raiz
        No(8, No(4, None, None), None)
        >>> t.insere(7)
        >>> t._raiz
        No(8, No(4, None, No(7, None, None)), None)
        '''
        # procura o local da inserção
        x = self._raiz
        # pai de x
        p = None
        while x != None:
            p = x
            if v < x.valor:
                x = x.esq
            else:
                x = x.dir
        # cria o nó que será inserido
        t = No(v, None, None)
        if p == None: # árvore estava vazia
            self._raiz = t
        elif v < p.valor:
            p.esq = t
        else:
            p.dir = t

    def _no_minimo(self, r):
        '''
        ArvoreBB, No -> No
        Devolve o nó que contém o valor mínimo da árvore com raiz r.
        Veja os exemplos de minimo.
        '''
        p = r
        if p == None:
            raise Exception('árvore vazia')
        while p.esq != None:
            p = p.esq
        return p

    def minimo(self):
        '''
        ArvoreBB -> Qualquer
        Devolve o valor mínimo da árvore.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t.minimo()
        3
        >>> t = ArvoreBB()
        >>> t.minimo()
        Traceback (most recent call last):
            ...
        Exception: árvore vazia
        '''
        p = self._no_minimo(self._raiz)
        return p.valor

    def _no_maximo(self, r):
        '''
        ArvoreBB, No -> No
        Devolve o nó que contém o valor máximo da árvore com raiz r.
        Veja os exemplos de maximo.
        '''
        p = r
        if p == None:
            raise Exception('árvore vazia')
        while p.dir != None:
            p = p.dir
        return p

    def maximo(self):
        '''
        ArvoreBB -> Qualquer
        Devolve o valor máximo da árvore
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t.maximo()
        15
        >>> t = ArvoreBB()
        >>> t.maximo()
        Traceback (most recent call last):
            ...
        Exception: árvore vazia
        '''
        p = self._no_maximo(self._raiz)
        return p.valor

    def altura(self):
        '''
        ArvoreBB -> Natural
        Calcula a altura da árvore.
        Exemplos:
        >>> t = ArvoreBB()
        >>> t.altura()
        -1
        >>> t.insere(3)
        >>> t.altura()
        0
        >>> t.insere(2)
        >>> t.altura()
        1
        >>> t = cria_arvore_teste()
        >>> t.altura()
        3
        '''
        # Para calcular a altura da árvore chamados a função auxiliar altura_no
        # com parâmetro self._raiz
        return altura_no(self._raiz)

    def __len__(self):
        '''
        No -> Natural
        Devolve a quantidade de nós na árvore.
        Exemplos:
        >>> t = ArvoreBB()
        >>> len(t)
        0
        >>> t.insere(3)
        >>> len(t)
        1
        >>> t.insere(4)
        >>> len(t)
        2
        >>> t = cria_arvore_teste()
        >>> len(t)
        7
        '''
        return tamanho_no(self._raiz)

    def remove(self, v):
        '''
        ArvoreBB, Qualquer -> None
        Remove o nó que contém o valor v.
        Lança uma exceção se v não está na árvore.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t.remove(3)
        >>> t
        ArvoreBB([4, 5, 6, 10, 12, 15])
        >>> t.remove(15)
        >>> t
        ArvoreBB([4, 5, 6, 10, 12])
        >>> t.insere(11)
        >>> t.insere(16)
        >>> t.insere(11.5)
        >>> t.remove(10)
        >>> t
        ArvoreBB([4, 5, 6, 11, 11.5, 12, 16])
        '''
        z = self._busca(v)
        if z == None:
            raise Exception(str(v) +' não encontrado')
        self._remove_no(z)

    def _remove_no(self, z):
        '''
        ArvoreBB, No -> None
        Remove o nó z da árvore.
        Veja os exemplos de remove.
        '''
        if z.esq == None:
            self._transplanta(z, z.dir)
        elif z.dir == None:
            self._transplanta(z, z.esq)
        else:
            y = self._no_minimo(z.dir)
            p = self.pai(y)
            if p != z:
                self._transplanta(y, y.dir)
                y.dir = z.dir
            self._transplanta(z, y)
            y.esq = z.esq

    def _transplanta(self, u, v):
        '''
        ArvoreBB, No, No -> None
        Altera o pai de u para ter como filho o v no lugar do u. Se u não tem
        pai, então v passa a ser a raiz da árvore.
        Veja os exemplos de remove.
        '''
        if self._raiz == u:
            self._raiz = v
        else:
            p = self.pai(u)
            if p.esq == u:
                p.esq = v
            else:
                p.dir = v

    def pai(self, u):
        '''
        ArvoreBB, No -> No
        Devolve o pai de u ou None se u não tem pai.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t.pai(t._busca(10)) == None
        True
        >>> t.pai(t._busca(5)).valor
        10
        >>> t.pai(t._busca(15)).valor
        10
        >>> t.pai(t._busca(6)).valor
        5
        '''
        p = self._raiz
        while p != None:
            if p.esq == u or p.dir == u:
                return p
            elif u.valor < p.valor:
                p = p.esq
            else:
                p = p.dir
        return None
    
    def __repr__(self):
        '''
        ArvoreBB -> String
        Devolve uma representação textual da árvore. Na representação os
        elementos aparecem em ordem.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t
        ArvoreBB([3, 4, 5, 6, 10, 12, 15])
        '''
        valores = str(self.em_ordem())
        return 'ArvoreBB(' + str(valores) + ')'

    def em_ordem(self):
        '''
        ArvoreBB -> Lista
        Devolve uma lista com os elementos da árvore visitados em ordem.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t.em_ordem()
        [3, 4, 5, 6, 10, 12, 15]
        '''
        valores = []
        def visita(p):
            '''
            No -> None
            Adiciona os elementos da árvore p em ordem na lista valores
            '''
            if p != None:
                visita(p.esq)
                valores.append(p.valor)
                visita(p.dir)
        visita(self._raiz)
        return valores

    def pre_ordem(self):
        '''
        ArvoreBB -> Lista
        Devolve uma lista com os elementos da árvore visitados em pré-ordem.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t.pre_ordem()
        [10, 5, 3, 4, 6, 15, 12]
        '''
        valores = []
        def visita(p):
            '''
            No -> None
            Adiciona os elementos da árvore p em pré-ordem na lista valores
            '''
            if p != None:
                valores.append(p.valor)
                visita(p.esq)
                visita(p.dir)
        visita(self._raiz)
        return valores
        
    def pos_ordem(self):
        '''
        ArvoreBB -> Lista
        Devolve uma lista com os elementos da árvore visitados em pós-ordem.
        Exemplos:
        >>> t = cria_arvore_teste()
        >>> t.pos_ordem()
        [4, 3, 6, 5, 12, 15, 10]
        '''
        valores = []
        def visita(p):
            '''
            No -> None
            Adiciona os elementos da árvore p em pós-ordem na lista valores
            '''
            if p != None:
                visita(p.esq)
                visita(p.dir)
                valores.append(p.valor)
        visita(self._raiz)
        return valores


if __name__ == '__main__':
    executa_testes()
